# -*- coding: utf-8 -*-

import sys
import json
from mamonsu.tools.zabbix_cli.request import Request


class Operations(object):

    _help_msg = """
Arguments must be:

mamonsu zabbix template list
mamonsu zabbix template show <template name>
mamonsu zabbix template id <template name>
mamonsu zabbix template delete <template id>
mamonsu zabbix template export <file>

mamonsu zabbix host list
mamonsu zabbix host show <host name>
mamonsu zabbix host id <host name>
mamonsu zabbix host delete <host id>
mamonsu zabbix host create <host name> <hostgroup id> <template id> <ip>
mamonsu zabbix host info templates <host id>
mamonsu zabbix host info hostgroups <host id>
mamonsu zabbix host info graphs <host id>
mamonsu zabbix host info items <host id>

mamonsu zabbix hostgroup list
mamonsu zabbix hostgroup show <hostgroup name>
mamonsu zabbix hostgroup id <hostgroup name>
mamonsu zabbix hostgroup delete <hostgroup id>
mamonsu zabbix hostgroup create <hostgroup name>

mamonsu zabbix item error <host name>
mamonsu zabbix item lastvalue <host name>
mamonsu zabbix item lastclock <host name>
"""

    def __init__(self, arg):

        self.arg = arg
        if len(self.arg.commands) < 2:
            self._print_help()

        self.req = Request(
            url='{0}/api_jsonrpc.php'.format(arg.zabbix_url),
            user=arg.zabbix_user,
            passwd=arg.zabbix_password)

        if self.arg.commands[0] == 'template':
            return self.template(self.arg.commands[1:])
        elif self.arg.commands[0] == 'hostgroup':
            return self.hostgroup(self.arg.commands[1:])
        elif self.arg.commands[0] == 'host':
            return self.host(self.arg.commands[1:])
        elif self.arg.commands[0] == 'item':
            return self.item(self.arg.commands[1:])
        else:
            self._print_help()

    def _print_help(self):
        sys.stderr.write(self._help_msg)
        sys.exit(1)

    def _generic_delete(self, typ, ids):
        if typ == 'template':
            params = [ids]
        elif typ == 'hostgroup':
            params = [ids]
        elif typ == 'host':
            params = [ids]
        else:
            sys.stderr.write('Unknown type: {0} for deleting'.format(typ))
            sys.exit(4)
        try:
            print(self.req.post(
                method='{0}.delete'.format(typ),
                params=params))
        except Exception as e:
            sys.stderr.write('List error: {0}\n'.format(e))
            sys.exit(3)

    def _generic_list(self, typ):
        if typ == 'template':
            name, fltr = 'name', 'host'
        elif typ == 'hostgroup':
            name, fltr = 'name', 'host'
        elif typ == 'host':
            fltr, name = 'host', 'host'
        else:
            sys.stderr.write('Unknown type: {0} for listing'.format(typ))
            sys.exit(4)
        try:
            for x in self.req.post(
                method='{0}.get'.format(typ),
                params={
                    'filter': {fltr: []}}):
                print(x[name])
        except Exception as e:
            sys.stderr.write('List error: {0}\n'.format(e))
            sys.exit(3)

    def _generic_show(self, typ, name, onlyid=False):
        if typ == 'template':
            fltr, ids = {'host': [name]}, 'templateid'
        elif typ == 'hostgroup':
            fltr, ids = {'name': [name]}, 'groupid'
        elif typ == 'host':
            fltr, ids = {'host': [name]}, 'hostid'
        else:
            sys.stderr.write('Unknown type: {0} for showing'.format(typ))
            sys.exit(4)
        try:
            x = self.req.post(
                method='{0}.get'.format(typ),
                params={'filter': fltr})
            if len(x) == 0:
                sys.stderr.write('{0} not found!\n'.format(name))
                sys.exit(2)
            if len(x) > 1:
                sys.stderr.write(
                    'Too many found: {0}\n'.format(x))
                sys.exit(2)
            if onlyid:
                print(x[0][ids])
            else:
                print(json.dumps(x[0], indent=2))
        except Exception as e:
            sys.stderr.write('Show error: {0}\n'.format(e))
            sys.exit(3)

    def _use_generic(self, args, typ):
        if args[0] == 'list':
            self._generic_list(typ)
            return True
        elif args[0] == 'show':
            if not len(args) == 2:
                return self._print_help()
            self._generic_show(typ, args[1])
            return True
        elif args[0] == 'id':
            if not len(args) == 2:
                return self._print_help()
            self._generic_show(typ, args[1], onlyid=True)
            return True
        elif args[0] == 'delete':
            if not len(args) == 2:
                return self._print_help()
            self._generic_delete(typ, args[1])
            return True
        else:
            return False

    def template(self, args):

        if self._use_generic(args, 'template'):
            return

        if args[0] == 'export':
            if not len(args) == 2:
                return self._print_help()
            file = args[1]
            try:
                if not self.req.post(
                    method='configuration.import',
                    params={
                        'format': 'xml',
                        'rules': {
                        'templates': {
                            'createMissing': True,
                            'updateExisting': True},
                        'applications': {
                            'createMissing': True,
                            'updateExisting': True},
                        'discoveryRules': {
                            'createMissing': True,
                            'updateExisting': True},
                        'graphs': {
                            'createMissing': True,
                            'updateExisting': True},
                        'items': {
                            'createMissing': True,
                            'updateExisting': True},
                        'triggers': {
                            'createMissing': True,
                            'updateExisting': True},
                        },
                        'source': open(file).read()}):
                    raise Exception('Export template error')
            except Exception as e:
                sys.stderr.write('Template export error: {0}\n'.format(e))
                sys.exit(3)
            return

        return self._print_help()

    def hostgroup(self, args):

        if self._use_generic(args, 'hostgroup'):
            return

        if args[0] == 'create':
            if not len(args) == 2:
                return self._print_help()
            name = args[1]
            try:
                result = self.req.post(
                    method='hostgroup.create',
                    params={'name': name})
                print(result['groupids'][0])
            except Exception as e:
                sys.stderr.write('Hostgroup create error: {0}\n'.format(e))
                sys.exit(3)
            finally:
                return

        return self._print_help()

    def host(self, args):

        def _extented_info(self, hostid, key, val):
            data = self.req.post(
                method='host.get',
                params={
                    'output': ['host'],
                    'hostids': hostid,
                    key: val
                })
            if not len(data) == 1:
                if len(data) == 0:
                    sys.stderr.write('Host not found: {0}\n'.format(hostid))
                else:
                    sys.stderr.write(
                        'Too many hosts found: {0}\n'.format(data))
                sys.exit(4)
            print(json.dumps(data[0], indent=2))

        if args[0] == 'info':
            if not len(args) == 3:
                return self._print_help()
            try:
                if args[1] == 'templates':
                    return _extented_info(
                        self, args[2],
                        'selectParentTemplates', ['templateid', 'name'])
                elif args[1] == 'hostgroups':
                    return _extented_info(
                        self, args[2],
                        'selectGroups', ['groupid', 'name'])
                elif args[1] == 'items':
                    return _extented_info(
                        self, args[2],
                        'selectItems', ['_key'])
                elif args[1] == 'graphs':
                    return _extented_info(
                        self, args[2],
                        'selectGraphs', ['graphid', 'name'])
                else:
                    self._print_help()
            except Exception as e:
                sys.stderr.write('Found error: {0}\n'.format(e))
                sys.exit(4)
            finally:
                return

        if self._use_generic(args, 'host'):
            return

        if args[0] == 'create':
            if not len(args) == 5:
                return self._print_help()
            name, groupid, templateid, ip = args[1], args[2], args[3], args[4]
            try:
                print(self.req.post(
                    method='host.create',
                    params={
                        'host': name,
                        'interfaces': [{
                            'type': 1, 'main': 1, 'useip': 1,
                            'ip': ip, 'dns': '', 'port': '10050'}],
                        'groups': [{'groupid': groupid}],
                        'templates': [{'templateid': templateid}]
                    }))
            except Exception as e:
                sys.stderr.write('Host create error: {0}\n'.format(e))
                sys.exit(4)
            finally:
                return

        return self._print_help()

    def item(self, args):
        if len(args) != 2:
            return self._print_help()
        typ, hostname = args[0], args[1]
        try:
            hosts = self.req.post('host.get', {'filter': {'host': [hostname]}})
            if len(hosts) == 0:
                sys.stderr.write('{0} not found!\n'.format(hostname))
                sys.exit(2)
            if len(hosts) > 1:
                sys.stderr.write('Too many found for {0}!\n'.format(hostname))
                sys.exit(2)
            host_id = hosts[0]['hostid']
            for item in self.req.post('item.get', {'hostids': [host_id]}):
                if typ == 'error':
                    if item['error'] == '':
                        continue
                print('{0}\t{1}'.format(
                    item['key_'],
                    item[typ]))
        except Exception as e:
            sys.stderr.write('Error find: {0}\n'.format(e))
            sys.exit(3)



# -*- coding: utf-8 -*-

import time
import logging
import traceback
import sys

from threading import Thread

from mamonsu.lib.const import Template


class PluginDisableException(Exception):
    pass


class Plugin(object):

    # plugin interval run
    Interval = 60

    # plugin config
    DEFAULT_CONFIG = {}  # type: Dict[str, str]

    _thread = None  # type: Thread
    _sender = False
    _enabled = True

    # for all childs
    is_child = True

    # const
    DELTA = Template.DELTA
    GRAPH_TYPE = Template.GRAPH_TYPE
    VALUE_TYPE = Template.VALUE_TYPE
    UNITS = Template.UNITS

    DELTA_SPEED = Template.DELTA.speed_per_second
    DELTA_CHANGE = Template.DELTA.simple_change

    def __init__(self, config):
        self.config = config
        self.log = logging.getLogger(
            self.__class__.__name__.upper())
        self.sender = None
        self.last_error_text = ''

        # from config => _plugin_config
        self._plugin_config = {}
        name = self.__class__.__name__.lower()
        if self.config.has_plugin_config(name):
            for x in self.config.plugin_options(name):
                self._plugin_config[x] = self.config.fetch(name, x)

    @classmethod
    def only_child_subclasses(self):
        plugins = []
        for klass in self.__subclasses__():
            if klass.is_child:
                plugins.append(klass)
            plugins.extend(klass.only_child_subclasses())
        return plugins

    @classmethod
    def set_default_config(cls, config):
        name = cls.__name__.lower()
        # if section already loaded via config file
        if not config.has_section(name) and len(cls.DEFAULT_CONFIG) > 0:
            config.add_section(name)
        for x in cls.DEFAULT_CONFIG:
            if config.has_option(name, x):
                continue
            value = cls.DEFAULT_CONFIG[x]
            if not isinstance(value, str):
                sys.stderr.write(
                    'Config value {0} in section {1} must'
                    ' be string! Fix plugin please.\n'.format(x, name))
            config.set(name, x, '{0}'.format(cls.DEFAULT_CONFIG[x]))

    # get value from config
    def plugin_config(self, name):
        if name not in self._plugin_config:
            return None
        return self._plugin_config[name]

    def start(self):
        self._thread = Thread(target=self._loop)
        self._thread.daemon = True
        self._thread.start()
        self.log.info('started ...')

    def is_alive(self):
        if self._thread is not None:
            return self._thread.is_alive()
        return False

    def run(self, sender):
        return None

    def is_sender(self):
        return self._sender

    def is_enabled(self):
        if self.plugin_config('enabled') == 'False':
            return False
        return self._enabled

    def disable(self):
        self._enabled = False

    def set_sender(self, sender):
        self.sender = sender

    def items(self, template):
        return None

    def graphs(self, template):
        return None

    def triggers(self, template):
        return None

    def discovery_rules(self, template):
        return None

    def _log_exception(self, e, trace):
        self.last_error_text = 'catch error: {0}'.format(e)
        self.log.error(self.last_error_text)
        self.log.info('hint: enable debug level to full exception trace')
        self.log.debug(trace)

    def _loop(self):
        while(True):
            last_start = time.time()
            try:
                self.run(self.sender)
            except PluginDisableException as e:
                text = 'disable plugin: {0}.'.format(e)
                self.log.info(text)
                return
            except Exception as e:
                trace = traceback.format_exc()
                self._log_exception(e, trace)
                return
            sleep_time = self.Interval - int(time.time() - last_start)
            if sleep_time > 0:
                time.sleep(sleep_time)
            else:
                self.log.error(
                    'Timeout: {0}s'.format(int(time.time() - last_start)))
                return



from .helpers import PerfData

from mamonsu.plugins.system.plugin import SystemPlugin as Plugin


class Cpu(Plugin):

    MaxPrivilegedTime = 60

    Items = [

        (r'\Processor(*)\% User Time', '[user_time]',
            'User time', Plugin.UNITS.percent, ('CC0000')),

        (r'\Processor(*)\% Idle Time', '[idle_time]',
            'Idle time', Plugin.UNITS.percent, ('00CC00')),

        (r'\Processor(*)\% Privileged Time', '[privileged_time]',
            'Privileged time', Plugin.UNITS.percent, ('770000')),

    ]

    def run(self, zbx):
        perf_services = []
        for _, item in enumerate(self.Items):
            perf_services.append(item[0])
        data = PerfData.get(perf_services, delay=1000)
        for idx, item in enumerate(self.Items):
            zbx.send('system.cpu{0}'.format(item[1]), float(data[idx]))

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'name': 'CPU {0}'.format(item[2].lower()),
                'key': 'system.cpu{0}'.format(item[1]),
                'units': item[3]
            })
        return result

    def graphs(self, template):
        items = []
        for item in self.Items:
            if item[4] is not None:
                items.append({
                    'key': 'system.cpu{0}'.format(item[1]),
                    'color': item[4]
                })
        graph = {'name': 'CPU overview', 'items': items, 'type': 1}
        return template.graph(graph)

    def triggers(self, template):
        return template.trigger({
            'name': 'CPU privileged time'
            'is too big on {HOSTNAME}',
            'expression': '{#TEMPLATE:system.cpu[privileged_time]'
            '.last()}&gt;' + str(self.MaxPrivilegedTime)
        })



# -*- coding: utf-8 -*-

import logging
import re
import time
from mamonsu.plugins.pgsql.pool import Pooler


class PostgresInfo(object):

    QueryCommonInfo = (
        """
select
    version(),
    (now() - pg_postmaster_start_time())::time,
    round(sum(blks_hit)*100/sum(blks_hit+blks_read), 2)
from
    pg_catalog.pg_stat_database """,
        ('version', 'uptime', 'cache hit')
    )

    QueryConnections = (
        """
select count(*), 'total' from pg_catalog.pg_stat_activity
union all
select count(*), 'active' from pg_catalog.pg_stat_activity
    where state = 'active'
union all
select count(*), 'idle' from pg_catalog.pg_stat_activity where state = 'idle'
union all
select count(*), 'idle in transaction' from pg_catalog.pg_stat_activity
    where state = 'idle in transaction'
"""
    )

    QueryRate = (
        """
select
    date_part('epoch', now()),
    sum(xact_commit),
    sum(xact_rollback)
from pg_catalog.pg_stat_database
"""
    )

    QueryPgSettings = (
        """
select
    name,
    setting,
    unit,
    context,
    vartype,
    source,
    boot_val,
    reset_val
from pg_catalog.pg_settings """,
        ('name', 'setting', 'unit', 'context',
            'vartype', 'source', 'boot_val', 'reset_val'),
        ({
            'Data safe': [
                'fsync', 'synchronous_commit',
                'archive_command', 'synchronous_standby_names'
            ],
            'Memory': [
                'shared_buffers', 'huge_pages', 'work_mem',
                'maintenance_work_mem', 'autovacuum_work_mem',
                'temp_buffers', 'max_connections'
            ],
            'Autovacuum': [
                'autovacuum', 'autovacuum_max_workers',
                'autovacuum_naptime', 'autovacuum_vacuum_cost_delay',
                'autovacuum_analyze_scale_factor',
                'autovacuum_analyze_threshold',
                'autovacuum_vacuum_scale_factor',
                'autovacuum_vacuum_threshold'
            ],
            'BgWriter': [
                'bgwriter_delay',
                'bgwriter_lru_maxpages',
                'bgwriter_lru_multiplier'
            ],
            'KeepAlive': [
                'tcp_keepalives_count',
                'tcp_keepalives_idle',
                'tcp_keepalives_interval'
            ],
            'Libraries': [
                'shared_preload_libraries',
                'session_preload_libraries',
                'local_preload_libraries'
            ]
        })
    )

    QueryDBList = (
        """
select
    d.datname,
    case when pg_catalog.has_database_privilege(d.datname, 'CONNECT')
         then pg_catalog.pg_size_pretty(pg_catalog.pg_database_size(d.datname))
         else 'No Access'
    end,
    pg_catalog.pg_get_userbyid(d.datdba),
    pg_catalog.pg_encoding_to_char(d.encoding),
    d.datcollate,
    d.datctype,
    pg_catalog.array_to_string(d.datacl, E'\n'),
    t.spcname,
    pg_catalog.shobj_description(d.oid, 'pg_database')
from pg_catalog.pg_database d
  join pg_catalog.pg_tablespace t on d.dattablespace = t.oid
order by 1 desc""",
        ('name', 'size', 'owner', 'encoding', 'collate',
            'ctype', 'privileges', 'tablespace', 'description')
    )

    BigTableInfo = (
        """
with big_tables as (
    select
        c.oid as relid,
        c.relname,
        n.nspname as schema,
        pg_catalog.pg_total_relation_size(c.oid) as size
    from pg_catalog.pg_class c
    left join pg_catalog.pg_namespace n ON n.oid = c.relnamespace
    where c.relkind in ('r','v','m','S','f','')
        order by pg_catalog.pg_total_relation_size(c.oid) desc limit 10)
select
    b.schema || '.' || b.relname as "table",
    pg_catalog.pg_size_pretty(b.size) as "size",
    s.idx_scan,
    s.seq_scan,

    case s.n_live_tup when 0 then 0
        else round(100*s.n_mod_since_analyze::float8/s.n_live_tup::float8) end as "since_analyze_perc",

    case s.n_live_tup when 0 then 0
        else round(100*s.n_dead_tup::float8/s.n_live_tup::float8) end as "dead_perc",

    case coalesce(io.heap_blks_read + io.heap_blks_hit, 0) when 0 then 0
        else round(100*io.heap_blks_hit::float8/(io.heap_blks_read+io.heap_blks_hit)::float8) end as "heap_hit_perc",

    case coalesce(io.idx_blks_read + io.idx_blks_hit, 0) when 0 then 0
        else round(100*io.idx_blks_hit::float8/(io.idx_blks_read + io.idx_blks_hit)::float8) end as "idx_hit_perc"

from
    pg_catalog.pg_stat_all_tables as s
    inner join pg_catalog.pg_statio_all_tables io on io.relid = s.relid
    inner join big_tables b on b.relid = s.relid
order by b.size desc
        """,
        ('table', 'size', 'idx scan', 'seq scan', '% since analyze',
            '% dead', '% heap', '% idx')
    )

    def __init__(self, args):
        self.args = args
        logging.info('Test connection...')
        self.connected = self._is_connection_work()
        if not self.connected:
            return
        logging.info('Collect pg common info...')
        self.common_info = self._collect_query(self.QueryCommonInfo)
        logging.info('Collect rate.,.')
        self.rate = self._collect_rate()
        logging.info('Collect connection info...')
        self.connections = self._collect_connections()
        logging.info('Collect pg_setting info...')
        self.settings = self._collect_query(self.QueryPgSettings)
        logging.info('Collect database list...')
        self.dblist = self._collect_query(self.QueryDBList)
        logging.info('Collect biggest table...')
        self.biggest_tables = self._collect_biggest()

    def collect(self):
        info = self.printable_info()
        logging.error("\n{0}\n".format(self.store_raw()))
        return info.encode('ascii', 'ignore').decode('ascii')

    def _is_connection_work(self):
        try:
            Pooler.query('select 1')
            return True
        except Exception as e:
            logging.error('Test query error: {0}'.format(e))
            return False

    _suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    def _humansize_bytes(self, nbytes):
        if nbytes == 0:
            return '0 B'
        i = 0
        while nbytes >= 1024 and i < len(self._suffixes) - 1:
            nbytes /= 1024.
            i += 1
        f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
        return '%s %s' % (f, self._suffixes[i])

    def _humansize(self, value):
        m = re.search('(\d+) (\S+)', value)
        if m is None:
            return value
        val, suff = m.group(1), m.group(2)
        val, suff = int(val), suff.upper()
        if suff == 'S':
            return value
        if suff == 'MS':
            return value
        if suff == 'B':
            return self._humansize_bytes(val)
        if suff == 'KB':
            return self._humansize_bytes(val * 1024)
        if suff == '4KB':
            return self._humansize_bytes(val * 1024 * 4)
        if suff == '8KB':
            return self._humansize_bytes(val * 1024 * 8)
        if suff == '16KB':
            return self._humansize_bytes(val * 1024 * 16)
        if suff == 'MB':
            return self._humansize_bytes(val * 1024 * 1024)
        if suff == 'GB':
            return self._humansize_bytes(val * 1024 * 1024 * 1024)
        if suff == 'TB':
            return self._humansize_bytes(val * 1024 * 1024 * 1024 * 1024)
        return value

    def store_raw(self):

        def format_obj(val):
            if isinstance(val, list) or isinstance(val, tuple):
                result = ''
                for row in val:
                    result += format_obj(row)
                result += "\n"
                return result
            if isinstance(val, dict):
                result = ''
                for key in val:
                    result += '{0}:{1}\n'.format(key, format_obj(val[key]))
                return result
            return '{0:10}|'.format(val)

        def format_out(info, val):
            return "# {0} ##################################\n{1}\n".format(
                info, val)

        if not self.connected:
            out = format_out('Test connection', 'Failed')
            return out
        out = format_out('PG COMMON', format_obj(self.common_info))
        out += format_out('PG RATE', format_obj(self.rate))
        out += format_out('PG CONNECTIONS', format_obj(self.connections))
        out += format_out('PG SETTINGS', format_obj(self.settings))
        out += format_out('PG DBLIST', format_obj(self.dblist))
        out += format_out('TABLES LIST', format_obj(self.biggest_tables))
        return out

    def printable_info(self):

        def format_header(info):
            return "\n###### {0:20} ###########################\n".format(info)

        def format_out(key, val):
            return "{0:40s}|    {1}\n".format(key, val)

        out = ''
        if not self.connected:
            out += format_header('PGSQL Error')
            out += format_out('Test connection', 'Failed')
            return out
        out += format_header('PostgreSQL')
        out += format_out('version', self.common_info[1][0])
        out += format_out('uptime', self.common_info[1][1])
        out += format_out('cache hit', '{0} %'.format(self.common_info[1][2]))
        out += format_out('tps', self.rate['_TPS'])
        out += format_out('rollbacks', self.rate['_ROLLBACKS'])
        out += format_header('Connections')
        for info in self.connections:
            count, name = info
            out += format_out(name, count)
        for key in self.QueryPgSettings[2]:
            out += format_header(key)
            for row in self.settings:
                for name in self.QueryPgSettings[2][key]:
                    if row[0] == name:
                        val = row[1]
                        if row[2] is not None:
                            val += ' {0}'.format(row[2])
                            val = self._humansize(val)
                        out += format_out(
                            name, val)
        out += format_header('Database sizes')
        for i, row in enumerate(self.dblist):
            if i == 0:
                continue
            out += format_out(row[0], self._humansize(row[1]))
        out += format_header('Biggest tables')
        big_table_header = ''
        for name in self.BigTableInfo[1][1:]:
            big_table_header = "{0}\t{1}".format(
                big_table_header, name)
        out += "{0:40s}|{1}\n".format('table', big_table_header)
        for i, key in enumerate(self.biggest_tables):
            out += format_out(
                key, self.biggest_tables[key])
        return out

    def _collect_query(self, query_desc):
        result = [query_desc[1]]
        try:
            for row in Pooler.query(query_desc[0]):
                result.append(row)
        except Exception as e:
            logging.error("Query {0} error: {1}".format(query_desc[0], e))
        return result

    def _collect_rate(self):
        result = {}
        result['_TPS'], result['_ROLLBACKS'] = '', ''
        try:
            result['row_1'] = Pooler.query(self.QueryRate)[0]
            time.sleep(2)
            result['row_2'] = Pooler.query(self.QueryRate)[0]
            exec_time = float(result['row_2'][0] - result['row_1'][0])
            result['_TPS'] = float(
                result['row_2'][1] - result['row_1'][1]) / exec_time
            result['_ROLLBACKS'] = float(
                result['row_2'][2] - result['row_1'][2]) / exec_time
        except Exception as e:
            logging.error('Query rate error: {0}'.format(e))
        return result

    def _collect_connections(self):
        result = []
        try:
            result = Pooler.query(self.QueryConnections)
        except Exception as e:
            logging.error('Query connections error: {0}'.format(e))
        return result

    def _collect_biggest(self):
        result = {}
        for info_dbs in Pooler.query('select datname \
                from pg_catalog.pg_database where datistemplate = false'):
            try:
                for info in Pooler.query(self.BigTableInfo[0], info_dbs[0]):
                    table_name = '{0}.{1}'.format(info_dbs[0], info[0])
                    result[table_name] = ''
                    for val in info[1:]:
                        result[table_name] = "{0}\t{1}".format(
                            result[table_name], val)
            except Exception as e:
                logging.error("Connect to db {0} error: {1}".format(
                    info_dbs[0], e))
        return result



__all__ = ['bgwriter', 'connections', 'databases']
__all__ += ['health', 'instance', 'xlog']
__all__ += ['pg_stat_statement', 'pg_buffercache', 'pg_wait_sampling']
__all__ += ['checkpoint', 'oldest', 'pg_locks']

from . import *



import ctypes
import mamonsu.lib.platform as platform

from ctypes import (Structure, Union, byref, c_double, c_longlong,
                    c_ulong, c_ulonglong, c_size_t, sizeof)
from ctypes.wintypes import HANDLE, LONG, LPCSTR, LPCWSTR, DWORD
from collections import namedtuple

HQUERY = HCOUNTER = HANDLE
psapi = ctypes.windll.psapi
pdh = ctypes.windll.pdh
kernel32 = ctypes.windll.kernel32

if platform.PY2:
    import codecs

    def u(x):
        return codecs.unicode_escape_decode(x)[0]
else:
    long = int

    def u(x):
        return x

PDH_FMT_RAW = long(16)
PDH_FMT_ANSI = long(32)
PDH_FMT_UNICODE = long(64)
PDH_FMT_LONG = long(256)
PDH_FMT_DOUBLE = long(512)
PDH_FMT_LARGE = long(1024)
PDH_FMT_1000 = long(8192)
PDH_FMT_NODATA = long(16384)
PDH_FMT_NOSCALE = long(4096)


class PerformanceInfo(Structure):
    _fields_ = [
        ('size', c_ulong),
        ('CommitTotal', c_size_t),
        ('CommitLimit', c_size_t),
        ('CommitPeak', c_size_t),
        ('PhysicalTotal', c_size_t),
        ('PhysicalAvailable', c_size_t),
        ('SystemCache', c_size_t),
        ('KernelTotal', c_size_t),
        ('KernelPaged', c_size_t),
        ('KernelNonpaged', c_size_t),
        ('PageSize', c_size_t),
        ('HandleCount', c_ulong),
        ('ProcessCount', c_ulong),
        ('ThreadCount', c_ulong),
    ]

    def __init__(self):
        self.size = sizeof(self)
        super(PerformanceInfo, self).__init__()

    @classmethod
    def get(cls):
        perfinfo = PerformanceInfo()
        psapi.GetPerformanceInfo(byref(perfinfo), perfinfo.size)
        return perfinfo


class DiskInfo(object):

    Types = {
        0: 'UNKNOWN',
        1: 'NO_ROOT_DIR',
        2: 'REMOVABLE',
        3: 'FIXED',
        4: 'REMOTE',
        5: 'CDROM',
        6: 'RAMDISK',
    }

    @classmethod
    def get_fixed_drivers(cls):
        drives = []
        bitmask = kernel32.GetLogicalDrives()
        for letter in u'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            if bitmask & 1:
                if kernel32.GetDriveTypeW(letter+':\\') == 3:
                    drives.append(letter)
            bitmask >>= 1
        return drives

    @classmethod
    def get_drive_info(cls, drive):
        _diskusage = namedtuple('disk_usage', 'total used free')
        used, total, free = c_ulonglong(), c_ulonglong(), c_ulonglong()
        drive = drive + u':\\'
        ret = kernel32.GetDiskFreeSpaceExW(
            str(drive), byref(used), byref(total), byref(free))
        if ret == 0:
            raise
        else:
            return _diskusage(total.value, used, free.value)


class PDH_Counter_Union(Union):
    _fields_ = [
        ('longValue', LONG),
        ('doubleValue', c_double),
        ('largeValue', c_longlong),
        ('ansiValue', LPCSTR),
        ('unicodeValue', LPCWSTR)
    ]


class PDHFmtCounterValue(Structure):
    _fields_ = [
        ('CStatus', DWORD),
        ('union', PDH_Counter_Union),
    ]


class PerfData(object):

    @classmethod
    def get(cls, counters, fmts=None, english=True, delay=0):

        if type(counters) is list:
            counters = [u(c) for c in counters]
        else:
            counters = [u(counters)]

        if fmts is None:
            fmts = []
            for counter in counters:
                fmts.append('double')

        getfmt = lambda fmt: globals().get(
            'PDH_FMT_' + fmt.upper(), PDH_FMT_LONG)
        if type(fmts) is list:
            ifmts = [getfmt(fmt) for fmt in fmts]
        else:
            ifmts = [getfmt(fmts)]
            fmts = [fmts]
        hQuery = HQUERY()
        hCounters = []
        values = []

        errs = pdh.PdhOpenQueryW(None, 0, byref(hQuery))
        if errs:
            raise(
                BaseException,
                'PdhOpenQueryW error: {0}'.format(errs))

        for counter in counters:
            hCounter = HCOUNTER()
            if english:
                errs = pdh.PdhAddEnglishCounterW(
                    hQuery, counter, 0, byref(hCounter))
            else:
                errs = pdh.PdhAddCounterW(
                    hQuery, counter, 0, byref(hCounter))
            if errs:
                raise(
                    BaseException,
                    'PdhAddCounterW error: {0}'.format(errs))
            hCounters.append(hCounter)

        errs = pdh.PdhCollectQueryData(hQuery)
        if errs:
            raise(
                BaseException,
                'PdhCollectQueryData error: {0}'.format(errs))
        if delay:
            kernel32.Sleep(delay)
            errs = pdh.PdhCollectQueryData(hQuery)
            if errs:
                raise(
                    BaseException,
                    'PdhCollectQueryData error: {0}'.format(errs))

        for i, hCounter in enumerate(hCounters):
            value = PDHFmtCounterValue()
            errs = pdh.PdhGetFormattedCounterValue(
                hCounter, ifmts[i], None, byref(value))
            if errs:
                raise(
                    BaseException,
                    'PdhGetFormattedCounterValue error: {0}'.format(errs))
            values.append(value)

        errs = pdh.PdhCloseQuery(hQuery)
        if errs:
            raise(
                BaseException,
                'PdhCloseQuery failed, error: {0}'.format(errs))

        return tuple([getattr(value.union, fmts[i] + 'Value')
                      for i, value in enumerate(values)])



# -*- coding: utf-8 -*-

from mamonsu.lib.const import Template


class ZbxTemplate(object):

    mainTemplate = u"""<?xml version="1.0" encoding="UTF-8"?>
<zabbix_export>
<version>2.0</version>
    <groups>
        <group>
            <name>Templates</name>
        </group>
    </groups>
    <templates>
        <template>
            <template>{template}</template>
            <name>{template}</name>
            <groups>
                <group>
                    <name>Templates</name>
                </group>
            </groups>
            <applications>
                <application>
                    <name>{application}</name>
                </application>
            </applications>
            <items>{items}</items>
            <discovery_rules>{discovery_rules}</discovery_rules>
            <macros>{macros}</macros>
        </template>
    </templates>
    <triggers>{triggers}</triggers>
    <graphs>{graphs}</graphs>
</zabbix_export>"""

# https://www.zabbix.com/documentation/2.0/manual/appendix/api/item/definitions
    item_defaults = [
        ('name', None), ('type', 7), ('snmp_community', None),
        ('multiplier', 0), ('inventory_link', 0),
        ('key', None), ('snmp_oid', None), ('history', 7),
        ('trends', 365), ('status', 0), ('delay', 60),
        ('value_type', Template.VALUE_TYPE.numeric_float),
        ('allowed_hosts', None), ('valuemap', None),
        ('units', Template.UNITS.none), ('delta', Template.DELTA.as_is),
        ('snmpv3_contextname', None), ('snmpv3_securityname', None),
        ('snmpv3_securitylevel', 0), ('snmpv3_authprotocol', 0),
        ('snmpv3_authpassphrase', None), ('snmpv3_privprotocol', 0),
        ('snmpv3_privpassphrase', None), ('formula', 1),
        ('delay_flex', None), ('params', None),
        ('ipmi_sensor', None), ('data_type', 0), ('authtype', 0),
        ('username', None), ('password', None), ('publickey', None),
        ('privatekey', None), ('port', None), ('description', None)
    ]

    trigger_defaults = [
        ('expression', None), ('name', None), ('url', None),
        ('status', 0), ('priority', 3), ('description', None),
        ('type', 0), ('dependencies', None)
    ]

    trigger_discovery_defaults = [
        ('expression', None), ('name', None), ('url', None),
        ('status', 0), ('priority', 3), ('description', None),
        ('type', 0)
    ]

    graph_values_defaults = [
        ('name', None), ('width', 900), ('height', 200),
        ('yaxismin', 0.0000), ('yaxismax', 100.0), ('show_work_period', 1),
        ('show_triggers', 1), ('type', 0), ('show_legend', 1),
        ('show_3d', 0), ('percent_left', 0.0), ('percent_right', 0.0),
        ('ymin_type_1', 0), ('ymax_type_1', 0), ('ymin_item_1', 0),
        ('ymax_item_1', 0)
    ]

    graph_items_defaults = [
        ('sortorder', None), ('drawtype', 0),
        ('color', '00CC00'), ('yaxisside', 0),
        ('calc_fnc', 2), ('type', Template.GRAPH_TYPE.normal)
    ]

    discovery_defaults = [
        ('name', None), ('type', 7), ('snmp_community', None),
        ('snmp_oid', None), ('delay', 60), ('status', 0),
        ('allowed_hosts', None), ('snmpv3_contextname', None),
        ('snmpv3_securityname', None), ('snmpv3_securitylevel', 0),
        ('snmpv3_authprotocol', 0), ('snmpv3_authpassphrase', None),
        ('snmpv3_privprotocol', 0), ('snmpv3_privpassphrase', None),
        ('delay_flex', None), ('params', None), ('filter', None),
        ('ipmi_sensor', None), ('authtype', 0),
        ('username', None), ('password', None), ('publickey', None),
        ('privatekey', None), ('port', None), ('lifetime', 7),
        ('description', None), ('key', None)
    ]

    def __init__(self, name, app):
        self.Application = app
        self.Template = name

    def xml(self, plugins=[]):
        # sort plugins!
        plugins.sort(key=lambda x: x.__class__.__name__)
        # create template
        template_data = {}
        template_data['template'] = self.Template
        template_data['application'] = self.Application
        template_data['items'] = self._get_all('items', plugins)
        template_data['macros'] = ''
        template_data['triggers'] = self._get_all('triggers', plugins)
        template_data['graphs'] = self._get_all('graphs', plugins)
        template_data['discovery_rules'] = self._get_all(
            'discovery_rules', plugins)
        return self.mainTemplate.format(**template_data)

    def _get_all(self, items='items', plugins=[]):
        result = ''
        for plugin in plugins:
            row = getattr(plugin, items)(self)
            if row is None:
                continue
            result += row
        return result

    def item(self, args={}, xml_key='item'):
        return '<{2}>{0}{1}</{2}>'.format(
            self._format_args(self.item_defaults, args),
            self._application(),
            xml_key)

    def trigger(self, args={}, xml_key='trigger', defaults=None):
        if defaults is None:
            defaults = self.trigger_defaults
        try:
            expression = args['expression']
        except KeyError:
            raise LookupError(
                'Miss expression in trigger: {0}.'.format(args))
        args['expression'] = expression.replace('#TEMPLATE', self.Template)
        return '<{1}>{0}</{1}>'.format(
            self._format_args(defaults, args),
            xml_key)

    def graph(self, args={}, xml_key='graph'):
        try:
            items = args['items']
        except KeyError:
            raise LookupError(
                'Miss item in graph: {0}.'.format(args))
        graph_items = ''
        for idx, item in enumerate(items):
            try:
                key = item['key']
            except KeyError:
                raise LookupError(
                    'Missed key in graph item: {0}.'.format(item))
            if 'sortorder' not in item:
                item['sortorder'] = idx
            row = '<graph_item>{0}<item><host>{1}'
            row += '</host><key>{2}</key></item></graph_item>'
            graph_items += row.format(
                self._format_args(self.graph_items_defaults, item),
                self.Template, key)
        result = '<{2}>{0}<graph_items>{1}</graph_items></{2}>'
        return result.format(
            self._format_args(self.graph_values_defaults, args),
            graph_items, xml_key)

    def discovery_rule(self, rule={}, items=[], triggers=[], graphs=[]):

        result_items = '<item_prototypes>'
        for item in items:
            result_items += self.item(item, xml_key='item_prototype')
        result_items += '</item_prototypes>'

        result_triggers = '<trigger_prototypes>'
        for trigger in triggers:
            result_triggers += self.trigger(
                trigger, xml_key='trigger_prototype',
                defaults=self.trigger_discovery_defaults)
        result_triggers += '</trigger_prototypes>'

        result_graphs = '<graph_prototypes>'
        for graph in graphs:
            result_graphs += self.graph(
                graph, xml_key='graph_prototype')
        result_graphs += '</graph_prototypes>'

        result = '<discovery_rule>{0}{1}{2}{3}</discovery_rule>'
        return result.format(
            self._format_args(self.discovery_defaults, rule),
            result_items, result_triggers, result_graphs)

    def _application(self):
        result = '<applications><application><name>{0}'
        result += '</name></application></applications>'
        return result.format(self.Application)

    def _format_args(self, defaults, override):
        result = ''
        for pair in defaults:
            key = pair[0]
            try:
                val = override[key]
            except KeyError:
                val = pair[1]
            if val is None:
                row = '<{0}/>'.format(key)
            else:
                row = '<{0}>{1}</{0}>'.format(key, val)
            result += row
        return result



import mamonsu.lib.platform as platform
from mamonsu.lib.plugin import Plugin

if platform.LINUX:
    import resource


class Health(Plugin):

    DEFAULT_CONFIG = {'max_memory_usage': str(40 * 1024 * 1024)}

    counter = 0

    def run(self, zbx):
        zbx.send('mamonsu.plugin.keepalive[]', 0)
        if platform.LINUX:
            usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
            zbx.send('mamonsu.memory.rss[max]', usage)
        self.counter += 1
        if self.counter > 9:
            self.log.info('=== Keep alive ===')
            self.counter = 0

    def items(self, template):
        # see supervisor.py:
        result = template.item({
            'name': 'Mamonsu: plugin errors',
            'key': 'mamonsu.plugin.errors[]',
            'value_type': Plugin.VALUE_TYPE.text  # text
        }) + template.item({
            'name': 'Mamonsu: plugin keep alive',
            'key': 'mamonsu.plugin.keepalive[]'
        })
        if platform.LINUX:
            result += template.item({
                'name': 'Mamonsu: rss memory max usage',
                'key': 'mamonsu.memory.rss[max]',
                'units': Plugin.UNITS.bytes
            })
        return result

    def triggers(self, template):
        result = template.trigger({
            'name': 'Mamonsu plugin errors '
            'on {HOSTNAME}. {ITEM.LASTVALUE}',
            'expression': '{#TEMPLATE:mamonsu.plugin.errors[].strlen()'
            '}&gt;1'
        }) + template.trigger({
            'name': 'Mamonsu nodata from {HOSTNAME}',
            'expression': '{#TEMPLATE:mamonsu.plugin.keepalive[]'
            '.nodata(180)}=1'
        })
        if platform.LINUX:
            result += template.trigger({
                'name': 'Mamonsu agent memory usage alert '
                'on {HOSTNAME}: {ITEM.LASTVALUE} bytes',
                'expression': '{#TEMPLATE:mamonsu.memory.rss[max].last()}'
                '&gt;' + self.plugin_config('max_memory_usage')
            })
        return result



from mamonsu.plugins.system.plugin import SystemPlugin as Plugin


class SystemUptime(Plugin):

    DEFAULT_CONFIG = {'uptime': str(60 * 5)}

    def run(self, zbx):
        uptime = open('/proc/uptime', 'r').read().split(' ')[0]
        zbx.send('system.uptime[]', int(float(uptime)))

    def items(self, template):
        return template.item({
            'name': 'System uptime',
            'key': 'system.uptime[]',
            'value_type': Plugin.VALUE_TYPE.numeric_unsigned,
            'units': Plugin.UNITS.uptime
        })

    def graphs(self, template):
        items = [{'key': 'system.uptime[]'}]
        graph = {'name': 'System uptime', 'items': items}
        return template.graph(graph)

    def triggers(self, template):
        return template.trigger({
            'name': 'System was restarted on '
            '{HOSTNAME} (uptime={ITEM.LASTVALUE})',
            'expression': '{#TEMPLATE:system.uptime[].last'
            '()}&lt;' + self.plugin_config('uptime')
        })



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class PgLocks(Plugin):

    Items = [
        # key, desc, color
        ('accessshare',
            'Read only queries',
            '0000CC'),
        ('rowshare',
            'SELECT FOR SHARE and SELECT FOR UPDATE',
            '00CC00'),
        ('rowexclusive',
            'Write queries',
            'CC0000'),
        ('shareupdateexclusive',
            'VACUUM, ANALYZE, CREATE INDEX CONCURRENTLY',
            'CC00CC'),
        ('share',
            'CREATE INDEX',
            '777777'),
        ('sharerowexclusive',
            'Locks from application',
            'CCCCCC'),
        ('exclusive',
            'Locks from application or some operation on system catalogs',
            'CCCC00'),
        ('accessexclusive',
            'ALTER TABLE, DROP TABLE, TRUNCATE, REINDEX, CLUSTER, '
            'VACUUM FULL, LOCK TABLE',
            '00CCCC')
    ]

    def run(self, zbx):
        result = Pooler.query("""
            select lower(mode), count(mode) FROM pg_catalog.pg_locks group by 1
            """)
        for row in result:
            for item in self.Items:
                if item[0] == row[0]:
                    zbx.send('pgsql.pg_locks[{0}]'.format(item[1]), row[1])

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'key': 'pgsql.pg_locks[{0}]'.format(item[0]),
                'name': 'PostgreSQL locks: {0}'.format(item[1]),
                'value_type': self.VALUE_TYPE.numeric_unsigned
            })
        return result

    def graphs(self, template):
        name, items = 'PostgreSQL locks sampling', []
        for item in self.Items:
            items.append({
                'key': 'pgsql.pg_locks[{0}]'.format(item[0]),
                'color': item[2]
            })
        return template.graph({'name': name, 'items': items})



from mamonsu.plugins.system.plugin import SystemPlugin as Plugin


class Net(Plugin):

    # position in line, key, desc, units
    Items = [
        (0, 'system.net.rx_bytes', 'RX bytes/s', Plugin.UNITS.bytes),
        (2, 'system.net.rx_errs', 'RX errors/s', Plugin.UNITS.none),
        (3, 'system.net.rx_drop', 'RX drops/s', Plugin.UNITS.none),
        (8, 'system.net.tx_bytes', 'TX bytes/s', Plugin.UNITS.bytes),
        (10, 'system.net.tx_errs', 'TX errors/s', Plugin.UNITS.none),
        (11, 'system.net.tx_drop', 'TX drops/s', Plugin.UNITS.none)
    ]

    def run(self, zbx):
        with open('/proc/net/dev', 'r') as f:
            devices = []
            for idx_line, line in enumerate(f, 1):
                if line.find(':') < 0 or line.find(' lo:') > 0 or idx_line < 1:
                    continue
                face, data = line.split(':')
                iface, values = face.strip(), [x for x in data.split()]
                for idx, value in enumerate(values):
                    for item in self.Items:
                        if item[0] == idx:
                            key = '{0}[{1}]'.format(item[1], iface)
                            zbx.send(key, float(value), self.DELTA_SPEED)
                devices.append({'{#NETDEVICE}': iface})
        zbx.send('system.net.discovery[]', zbx.json({'data': devices}))

    def discovery_rules(self, template):
        items = []
        for item in self.Items:
            items.append({
                'key': item[1]+'[{#NETDEVICE}]',
                'name': 'Network device {#NETDEVICE}: ' + item[2],
                'units': item[3]})
        rule = {
            'name': 'Net iface discovery',
            'key': 'system.net.discovery[]',
            'filter': '{#NETDEVICE}:.*'
        }
        graphs = [{
            'name': 'Network device: {#NETDEVICE}',
            'items': [{
                    'color': 'CC0000',
                    'key': 'system.net.rx_bytes[{#NETDEVICE}]'},
                {
                    'color': '0000CC',
                    'key': 'system.net.tx_bytes[{#NETDEVICE}]'}]
        }]
        return template.discovery_rule(rule=rule, items=items, graphs=graphs)



# -*- coding: utf-8 -*-

import os
import logging

import mamonsu.lib.platform as platform
from mamonsu.tools.sysinfo.linux import SysInfoLinux


class AutoTuneSystem(SysInfoLinux):

    _SYSCTL_FILE = '/etc/sysctl.conf'
    _SYSCTL_MAGIC_LINE = "#### mamonsu auto tune ####\n"

    def __init__(self, args):

        sudo = False if (args.disable_sudo is False) else True
        super(AutoTuneSystem, self).__init__(use_sudo=sudo)

        self.args = args
        self._sysctl_data = []

        self._dirty()
        self._min_free()

        self._write_sysctl_file()

    # dirty bytes
    def _dirty(self):
        if not platform.LINUX:
            return
        total = self.meminfo['_TOTAL']
        # if total < 1Gb, dont tune dirty bytes
        if total > 1024 * 1024 * 1024:
            self._add_sysctl('vm.dirty_background_bytes', 32 * 1024 * 1024)
            self._add_sysctl('vm.dirty_bytes', 64 * 1024 * 1024)

    # defragment
    def _min_free(self):
        if not platform.LINUX:
            return
        total = self.meminfo['_TOTAL']
        self._add_sysctl('vm.min_free_kbytes', int(5 * total / (100 * 1024)))

    def _add_sysctl(self, key, value):
        self._sysctl_data.append("{0} = {1}\n".format(key, value))

    # write to sysctl
    def _write_sysctl_file(self):
        if not platform.LINUX:
            return
        if not os.path.isfile(self._SYSCTL_FILE):
            return
        try:
            all_sysctl = []
            with open(self._SYSCTL_FILE, 'r') as f:
                for line in f:
                    if self._SYSCTL_MAGIC_LINE == line:
                        break
                    all_sysctl.append(line)
            all_sysctl.append("\n{0}".format(self._SYSCTL_MAGIC_LINE))
            all_sysctl += self._sysctl_data
            if self.args.dry_run:
                logging.info('dry run (write sysctl vars: {0}):\n{1}'.format(
                    self._SYSCTL_FILE, ''.join(all_sysctl)))
                return
            with open(self._SYSCTL_FILE, 'w') as f:
                f.write(''.join(all_sysctl))
                f.flush()
        except Exception as e:
            logging.error('Write to sysctl error: {0}'.format(e))



from .helpers import DiskInfo, PerfData

from mamonsu.plugins.system.plugin import SystemPlugin as Plugin


class DiskStats(Plugin):

    def run(self, zbx):
        devices = []
        for disk in DiskInfo.get_fixed_drivers():
            perf_services = [
                r'\LogicalDisk({0}:)\Disk Reads/sec'.format(disk),
                r'\LogicalDisk({0}:)\Disk Writes/sec'.format(disk),
                r'\LogicalDisk({0}:)\% Idle Time'.format(disk),
                r'\LogicalDisk({0}:)\Avg. Disk Queue Length'.format(disk)]
            data = PerfData.get(perf_services, delay=1000)
            zbx.send('system.disk.read[{0}]'.format(disk), float(data[0]))
            zbx.send('system.disk.write[{0}]'.format(disk), float(data[1]))
            zbx.send('system.disk.idle[{0}]'.format(disk), float(data[2]))
            zbx.send('system.disk.queue_avg[{0}]'.format(disk), float(data[3]))
            devices.append({'{#LOGICALDEVICE}': disk})
            del perf_services, data
        zbx.send('system.disk.discovery[]', zbx.json({'data': devices}))
        del devices

    def discovery_rules(self, template):
        rule = {
            'name': 'Logical disks discovery',
            'key': 'system.disk.discovery[]',
            'filter': '{#LOGICALDEVICE}:.*'
        }
        items = [
            {
                'key': 'system.disk.read[{#LOGICALDEVICE}]',
                'name': 'Logical device {#LOGICALDEVICE}: read op/sec',
                'delta': Plugin.DELTA.speed_per_second},
            {
                'key': 'system.disk.write[{#LOGICALDEVICE}]',
                'name': 'Logical device {#LOGICALDEVICE}: write op/sec',
                'delta': Plugin.DELTA.speed_per_second},
            {
                'key': 'system.disk.queue_avg[{#LOGICALDEVICE}]',
                'name': 'Logical device {#LOGICALDEVICE}: queue'},
            {
                'key': 'system.disk.idle[{#LOGICALDEVICE}]',
                'name': 'Logical device {#LOGICALDEVICE}: idle time (%)',
                'units': Plugin.UNITS.percent}]
        graphs = [{
            'name': 'Logical devices overview: {#LOGICALDEVICE}',
            'items': [{
                    'color': 'CC0000',
                    'key': 'system.disk.read[{#LOGICALDEVICE}]'},
                {
                    'color': '0000CC',
                    'key': 'system.disk.write[{#LOGICALDEVICE}]'},
                {
                    'key': 'system.disk.queue_avg[{#LOGICALDEVICE}]',
                    'name': 'Logical device {#LOGICALDEVICE}: queue',
                    'yaxisside': 1}]
        }]
        return template.discovery_rule(rule=rule, items=items, graphs=graphs)



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class PgStatStatement(Plugin):

    # zbx_key, sql, desc, unit, delta, (Graph, color, side)
    Items = [

        ('stat[read_bytes]',
            'sum(shared_blks_read+local_blks_read+temp_blks_read)*8*1024',
            'read bytes/s', Plugin.UNITS.bytes, Plugin.DELTA.speed_per_second,
            ('PostgreSQL statements: bytes', 'BBBB00', 0)),
        ('stat[write_bytes]',
            'sum(shared_blks_written+local_blks_written'
            '+temp_blks_written)*8*1024',
            'write bytes/s', Plugin.UNITS.bytes, Plugin.DELTA.speed_per_second,
            ('PostgreSQL statements: bytes', '00CC00', 0)),
        ('stat[dirty_bytes]',
            'sum(shared_blks_dirtied+local_blks_dirtied)*8*1024',
            'dirty bytes/s', Plugin.UNITS.bytes, Plugin.DELTA.speed_per_second,
            ('PostgreSQL statements: bytes', '0000CC', 0)),

        ('stat[read_time]',
            'sum(blk_read_time)/float4(100)',
            'read io time', Plugin.UNITS.s, Plugin.DELTA.speed_per_second,
            ('PostgreSQL statements: spend time', '00CC00', 0)),
        ('stat[write_time]',
            'sum(blk_write_time)/float4(100)',
            'write io time', Plugin.UNITS.s, Plugin.DELTA.speed_per_second,
            ('PostgreSQL statements: spend time', '0000CC', 0)),
        ('stat[other_time]',
            'sum(total_time-blk_read_time-blk_write_time)/float4(100)',
            'other (mostly cpu) time',
            Plugin.UNITS.s, Plugin.DELTA.speed_per_second,
            ('PostgreSQL statements: spend time', 'BBBB00', 0))
    ]

    def run(self, zbx):
        if not self.extension_installed('pg_stat_statements'):
            return
        params = [x[1] for x in self.Items]
        result = Pooler.query('\
            select {0} from public.pg_stat_statements'.format(
            ', '.join(params)))
        for idx, val in enumerate(result[0]):
            key, val = 'pgsql.{0}'.format(
                self.Items[idx][0]), int(val)
            zbx.send(key, val, self.Items[idx][4])

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'key': 'pgsql.{0}'.format(item[0]),
                'name': 'PostgreSQL statements: {0}'.format(item[2]),
                'value_type': self.VALUE_TYPE.numeric_float,
                'units': item[3]})
        return result

    def graphs(self, template):
        all_graphs = [
            ('PostgreSQL statements: bytes', None),
            ('PostgreSQL statements: spend time', 1)]
        result = ''
        for graph_item in all_graphs:
            items = []
            for item in self.Items:
                if item[5][0] == graph_item[0]:
                    items.append({
                        'key': 'pgsql.{0}'.format(item[0]),
                        'color': item[5][1],
                        'yaxisside': item[5][2]})
            # create graph
            graph = {'name': graph_item[0], 'items': items}
            if graph_item[1] is not None:
                graph['type'] = graph_item[1]
            result += template.graph(graph)
        return result
